/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.processor.processor;

import java.util.Collection;
import java.util.Collections;
import java.util.Optional;
import java.util.Set;
import java.util.stream.Stream;

import org.apache.commons.lang3.StringUtils;
import org.springframework.beans.factory.annotation.Qualifier;
import org.springframework.boot.autoconfigure.condition.ConditionalOnProperty;
import org.springframework.http.HttpEntity;
import org.springframework.http.HttpStatusCode;
import org.springframework.http.MediaType;
import org.springframework.http.client.MultipartBodyBuilder;
import org.springframework.stereotype.Service;
import org.springframework.util.MultiValueMap;
import org.springframework.web.reactive.function.client.ClientResponse;
import org.springframework.web.reactive.function.client.WebClient;

import de.ozgcloud.common.errorhandling.TechnicalException;
import de.ozgcloud.processor.ProcessorManagerConfiguration;
import de.ozgcloud.processor.processor.ProcessorProperties.Form;
import de.ozgcloud.processor.processor.ProcessorProperties.Processor;
import de.ozgcloud.processor.result.ProcessorTechnicalException;
import de.ozgcloud.processor.vorgang.Vorgang;
import lombok.RequiredArgsConstructor;
import lombok.extern.log4j.Log4j2;
import reactor.core.publisher.Mono;

@ConditionalOnProperty(prefix = "ozgcloud.processors[0]", name = "address")
@Service(ProcessorManagerConfiguration.PROCESSOR_SERVICE_NAME) // NOSONAR
@RequiredArgsConstructor
@Log4j2
public class ProcessorService {

	private static final String KEY_VORGANG = "vorgang";

	@Qualifier(ProcessorManagerConfiguration.PROCESSOR_PROPERTIES_NAME) // NOSONAR
	private final ProcessorProperties properties;

	private final WebClient webClient;

	public Stream<Mono<ProcessorResult>> processVorgang(Vorgang vorgang) {
		return getProcessors(vorgang).map(processor -> callProcessor(processor, vorgang));
	}

	Stream<Processor> getProcessors(Vorgang vorgang) {
		return properties.getProcessors().parallelStream().filter(processor -> hasFormEngineNameAndFormIds(processor.getForms(), vorgang));
	}

	boolean hasFormEngineNameAndFormIds(Collection<Form> forms, Vorgang vorgang) {
		return Optional.ofNullable(forms).orElse(Collections.emptySet()).stream()
				.filter(form -> StringUtils.equals(form.getFormEngineName(), vorgang.getFormEngineName()))
				.anyMatch(form -> StringUtils.equals(form.getFormId(), vorgang.getFormId()));
	}

	public Stream<Mono<ProcessorResult>> processVorgang(Vorgang vorgang, Collection<String> processorNames) {
		return getProcessors(processorNames)
				.filter(processor -> containsMatchingForm(vorgang, processor.getForms()))
				.map(processor -> callProcessor(processor, vorgang));
	}

	private boolean containsMatchingForm(Vorgang vorgang, Set<Form> forms) {
		return forms.stream().anyMatch(form -> StringUtils.equals(vorgang.getFormEngineName(), form.getFormEngineName())
				&& StringUtils.equals(vorgang.getFormId(), form.getFormId()));
	}

	Stream<Processor> getProcessors(Collection<String> processorNames) {
		return properties.getProcessors().stream().filter(processor -> processorNames.contains(processor.getName()));
	}

	Mono<ProcessorResult> callProcessor(Processor processor, Vorgang vorgang) {
		LOG.info("Sending Vorgang {} to processors.", vorgang.getId());
		return webClient.post()
				.uri(processor.getAddress())
				.bodyValue(createRequestBody(vorgang))
				.retrieve()
				.onStatus(HttpStatusCode::is3xxRedirection, this::buildRedirectError)
				.bodyToMono(ProcessorResult.class)
				.onErrorMap(exception -> new ProcessorTechnicalException(vorgang.getId(), processor.getName(), "Failed to request Vorgang processor.",
						exception))
				.doOnSuccess(processorResult -> processorResult.toBuilder().processorName(processor.getName()).build());
	}

	MultiValueMap<String, HttpEntity<?>> createRequestBody(Vorgang vorgang) { // NOSONAR
		var multipartBodyBuilder = new MultipartBodyBuilder();
		multipartBodyBuilder.part(KEY_VORGANG, vorgang, MediaType.APPLICATION_JSON);
		return multipartBodyBuilder.build();
	}

	Mono<Throwable> buildRedirectError(ClientResponse clientResponse) {
		return Mono.error(new TechnicalException("Resource was moved (%s) ".formatted(clientResponse.statusCode())));
	}
}